"""Uses the argparse module to parse arguments/flags given on
the commandline and returns them as a Dictionary."""

import argparse


def parse_arguments():
    """Read arguments from the command line.

    parse_arguments() uses Python's agparser to read arguments from the command
    line. It also sets defaults and provides help and hints abouth which
    arguments are available

    Availble arguments:
    -h, --help: Automatically generated, prints options for help

    -c, --catch-up: Catch up to the current state of the timeline without
    tooting anything. This is useful if the script has not been running for
    a while and would otherwise (re)post lots of old group-toots.

    -d, --dry-run: Parse new messages but do not upload or toot anything.
    Shows the ID of notifications it would have processed instead.

    -g, --group: group handle the script is currently running for. Needed
    by configparser to find its configuration and can be chosen freely.

    -k, --ketchup: Same as -c or --catch-up, for the sake of lol.

    --version: Show tootgroup.py version and exit"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--catch-up",
        action="store_true",
        help="Catch up to the current state of the timeline without tooting "
        + "anything. This is useful if the script has not been running for a "
        + "while and would otherwise (re)post lots of old group-toots.",
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help="Parse new messages but do not upload or toot anything. "
        + "Shows the ID of notifications it would have processed instead.",
    )
    parser.add_argument(
        "-g",
        "--group",
        default="default",
        help="Input a handle for the Mastodon account. tootgroup.py stores all "
        + "information connected to a specific group account under this name. "
        + "By choosing different handles it is possible to manage multiple "
        + "Mastodon groups from the same skript. This can be chosen freely but "
        + " it is wise to use a name that is related to the Mastodon account "
        + " it will be used with. If no handle is given, "
        + '"%(default)s" is always used instead.',
    )
    parser.add_argument(
        "-k",
        "--ketchup",
        action="store_true",
        help="Same as -c or --catch-up for the sake of lol!",
    )
    parser.add_argument(
        "--version", action="store_true", help="Show tootgroup.py version and exit."
    )
    args = parser.parse_args()
    arguments = {}
    arguments["group_name"] = args.group
    arguments["catch_up"] = False
    arguments["dry_run"] = False
    arguments["show_version"] = False
    if args.catch_up or args.ketchup:
        arguments["catch_up"] = True
    if args.dry_run:
        arguments["dry_run"] = True
    if args.version:
        arguments["show_version"] = True

    return arguments



"""Manages the applications config file storage location, reading
and writing."""

import configparser
import os
import sys

import appdirs
import mastodon


def new_credentials_from_server(config_store, config):
    """Register tootgroup.py at a Mastodon (or Pleroma) server and get
    user credentials.

    "config_store" dictionary containting config file name and path as
    well as the current group's name

    "config" the configuration as read in from configparser

    This will be run if tootgroup.py is started for the first time, if its
    configuration files have been deleted or if some elements of the
    configuration are missing.
    """
    group_name = config_store["group_name"]

    # Register tootgroup.py app at the Mastodon server
    try:
        mastodon.Mastodon.create_app(
            "tootgroup.py",
            api_base_url=config[group_name]["mastodon_instance"],
            to_file=config_store["directory"] + config[group_name]["client_id"],
        )
        # Create Mastodon API instance
        masto = mastodon.Mastodon(
            client_id=config_store["directory"] + config[group_name]["client_id"],
            api_base_url=config[group_name]["mastodon_instance"],
        )
    except Exception as ex:
        print("")
        print("\n##################################################################")
        print("The Mastodon instance URL is wrong or the server does not respond.")
        print("See the error message for more details:")
        print(ex)
        print("")
        print("tootgroup.py will exit now. Run it again to try once more!")
        print("##################################################################\n")
        sys.exit(0)

    # Log in once with username and password to get an access token for future logins.
    # Ask until login succeeds or at most 3 times before the skript gives up.
    i = 0
    while i < 3:
        i += 1
        try:
            masto.log_in(
                input("Username (e-Mail) to log into Mastodon Instance: "),
                input("Password: "),
                to_file=config_store["directory"] + config[group_name]["access_token"],
            )
            break
        except Exception:
            print("\nUsername and/or Password did not match!")
            if i < 3:
                print("Please enter them again.\n")
            else:
                print("tootgroup.py will exit now. Run it again to try once more!\n")
                sys.exit(0)


def parse_configuration(config_store):
    """Read configuration from file, handle first-run situations and errors.

    "config_store" dictionary containting config file name and path as
    well as the current group's name

    parse_configuration() uses Python's configparser to read and interpret the
    config file. It will detect a missing config file or missing elements and
    then try to solve problems by asking the user for more information. This
    does also take care of a first run situation where nothing is set up yet and
    in that way act as an installer!

    parse_configuration should always return a complete and usable configuration"""
    config = configparser.ConfigParser()
    config.read(config_store["directory"] + config_store["filename"])

    group_name = config_store["group_name"]
    get_new_credentials = False

    # Is there already a section for the current tootgroup.py
    # group. If not, create it now.
    if not config.has_section(group_name):
        config[group_name] = {}
        print(
            'Group "'
            + group_name
            + '" not in configuration. '
            + "- Setting it up now..."
        )
        config_store["write_NEW"] = True
        config_store["first_run"] = True

    # Do we have a mastodon instance URL? If not, we have to
    # ask for it and register with our group's server first.
    if not config.has_option(group_name, "mastodon_instance"):
        config[group_name]["mastodon_instance"] = ""
        print("We need a Mastodon server to connect to!")
    if config[group_name]["mastodon_instance"] == "":
        config[group_name]["mastodon_instance"] = input(
            "Enter the "
            "URL of the Mastodon instance your group account is "
            "running on: "
        )
        get_new_credentials = True
        config_store["write_NEW"] = True

    # Where can the client ID be found and does the file exist?
    # If not, re-register the client.
    if not config.has_option(group_name, "client_id"):
        config[group_name]["client_id"] = ""
    if config[group_name]["client_id"] == "":
        config[group_name]["client_id"] = group_name + "_clientcred.secret"
        get_new_credentials = True
        config_store["write_NEW"] = True
    if not os.path.isfile(config_store["directory"] + config[group_name]["client_id"]):
        get_new_credentials = True

    # Where can the user access token be found and does the file exist?
    # If not, re-register the client to get new user credentials
    if not config.has_option(group_name, "access_token"):
        config[group_name]["access_token"] = ""
    if config[group_name]["access_token"] == "":
        config[group_name]["access_token"] = group_name + "_usercred.secret"
        get_new_credentials = True
        config_store["write_NEW"] = True
    if not os.path.isfile(
        config_store["directory"] + config[group_name]["access_token"]
    ):
        get_new_credentials = True

    # Should tootgroup.py accept public mentions for retooting?
    while config[group_name].get("accept_retoots") not in ("yes", "no"):
        config[group_name]["accept_retoots"] = input(
            "\nShould tootgroup.py RETOOT public mentions "
            + "from group members? [ yes | no ]: "
        ).lower()
        config_store["write_NEW"] = True

    # Should tootgroup.py accept direct messages for reposting?
    while config[group_name].get("accept_dms") not in ("yes", "no"):
        config[group_name]["accept_dms"] = input(
            "\nShould tootgroup.py REPOST direct messages from group "
            + "members? [ yes | no ]: "
        ).lower()
        config_store["write_NEW"] = True

    # How public should the toots from direct messages be?
    while config[group_name].get("dm_visibility") not in (
        "private",
        "unlisted",
        "public",
    ):

        # This option was added in tootgroup.py v1.2 - auto-upgrade
        # existing installations as to not require manual intervention
        if not config_store["first_run"]:
            print("")
            print("### Automatic configuration update! ###")
            print(
                "Visibility of REPOSTS will be set to 'public' as this is standard behaviour."
            )
            print("")
            config[group_name]["dm_visibility"] = "public"

        else:
            config[group_name]["dm_visibility"] = input(
                "\nWhat visibility should the toots created from DMs "
                + "have? Unlisted is recommended for testing, public for "
                + "regular use.\n[ private | unlisted | public ]: "
            ).lower()

        config_store["write_NEW"] = True

    # The ID of the last group-toot is needed to check for newly arrived
    # notifications and will be persisted here. It is initially set up with
    # "catch-up" to indicate this situation.
    # Tootgroup.py will then get sane values and continue.
    if (not config.has_option(group_name, "last_seen_id")) or (
        config[group_name]["last_seen_id"] == ""
    ):
        config[group_name]["last_seen_id"] = "catch-up"
        config_store["write_NEW"] = True

    # Some registration info or credentials were missing - we have to register
    # tootgroup.py with our Mastodon server instance. (again?)
    if get_new_credentials:
        print("Some credentials are missing, need to get new ones...")
        new_credentials_from_server(config_store, config)

    # Have there been any changes to the configuration?
    # If yes we have to write them to the config file
    if config_store["write_NEW"]:
        write_configuration(config_store, config)

    # Configuration should be complete and working now - return it.
    return config


def setup_configuration_store():
    """Search for the configuration file location. Start with the
    skript's home directory and try the OS specific user configuration
    directory next. Local takes precedence but if no configuration
    is found, the new one will be created in the OS specific user
    config path if possible. If that fails, the local directory will
    again be used as a last resort. This is to ensure
    backwards-compatibility and to enable local overrides for
    development/testing.

    Returns a dictinary containing  basic information about how to
    find the configuration. Some dictionary values will be populated
    later (see below for available fields)"""

    config_store = {
        "filename": "",
        "directory": "",
        "group_name": "",
        "first_run": False,
        "write_NEW": False,
    }

    # Get name and path of the running application. It is used to
    # determine its configuration storage path
    stack = sys._getframe(1).f_globals  # get caller's globals
    application_namepath = stack["__file__"]
    application_path, application_name = os.path.split(application_namepath)
    application_path = application_path + "/"

    # The config file is named after the application with
    # the extension is changed from ".py" to ".conf"
    # Only the last occurence of ".py" is replaced!
    old_ext = ".py"
    new_ext = ".conf"
    config_filename = new_ext.join(application_name.rsplit(old_ext, 1))
    config_store["filename"] = config_filename

    local_path = application_path
    os_user_config_path = appdirs.AppDirs(application_name).user_data_dir + "/"
    config_store["directory"] = local_path

    # Is there a config file in the tootgroup.py directory?
    if os.path.isfile(local_path + config_filename):
        print("Found local configuration and using it...")

    # Is there a config file in the system's user config directory?
    elif os.path.isfile(os_user_config_path + config_filename):
        config_store["directory"] = os_user_config_path
        # This is the default, do not output anything

    # Did not find any config file, default to user config dir!
    else:
        config_store["directory"] = os_user_config_path
        config_store["write_NEW"] = True
        config_store["first_run"] = True
        # Create config dir if it does not exist yet
        try:
            os.makedirs(config_store["directory"], exist_ok=True)
            print("No configuration found!")
            print(
                "tootgroup.py will continue to run but first "
                + "ask for all information it requires...\n"
                + "Configuration will then be stored under "
                + config_store["directory"]
                + "\n"
            )
        except Exception:
            # Cannot create config directory, have to use the
            # tootgroup.py dir instead
            config_store["directory"] = local_path
            print("No configuration found!")
            print(
                "tootgroup.py will continue to run but first "
                + "ask for all information it requires...\n"
                + "Configuration will then be stored under "
                + config_store["directory"]
                + "\n"
            )

    return config_store


def write_configuration(config_store, config):
    """Write out the configuration into the config file.

    "config_store: dictionary containting config file name and path

    "config" configparser object containing the current configuration.

    This can be called whenever the configuration needs to be persisted by
    writing it to the disk."""
    try:
        with open(
            config_store["directory"] + config_store["filename"], "w", encoding="utf-8"
        ) as configfile:
            config.write(configfile)

    except Exception as ex:
        print("")
        print("\n############################################################")
        print("Cannot write to configuration file:")
        print(ex)
        print("Try to fix the problem and run toogroup.py again afterwards.")
        print("############################################################\n")
        sys.exit(0)



"""Current version of tootgroup.py - Kept in a single place
for easier modification before every new release."""

CURRENT = "1.4.1"


def print_version():
    """Do just that"""
    print("tootgroup.py version: " + CURRENT)



#!/usr/bin/env python3
"""tootgroup.py

Andreas Schreiner
@andis@chaos.social
andreas.schreiner@sonnenmulde.at

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

See attached LICENSE file.
"""
import html
import os
import re
import sys
import tempfile

import mastodon
import requests

import tootgroup_tools

TOOTGROUP_VERSION = tootgroup_tools.version.CURRENT


def main():
    """Execution starts here"""

    # Read commandline arguments and flags from input
    commandline_arguments = tootgroup_tools.commandline_arguments.parse_arguments()

    # if the "--version" argument has been given, show version and exit.
    if commandline_arguments["show_version"]:
        tootgroup_tools.version.print_version()
        sys.exit(0)

    # Get the configuration storage location
    config_store = tootgroup_tools.configuration_management.setup_configuration_store()

    # Get the Mastodon/Pleroma account handle the script is running for.
    config_store["group_name"] = commandline_arguments["group_name"]
    group_name = config_store["group_name"]

    # Get and validate configuration from the config file.
    my_config = tootgroup_tools.configuration_management.parse_configuration(
        config_store
    )

    # If the "catch-up" commandline argument has been given, reset the last seen
    # ID so that tootgroup.py only updates its according config value without
    # retooting anything.
    if commandline_arguments["catch_up"]:
        my_config[group_name]["last_seen_id"] = "catch-up"

    # Create Mastodon API instance.
    masto = mastodon.Mastodon(
        client_id=config_store["directory"] + my_config[group_name]["client_id"],
        access_token=config_store["directory"] + my_config[group_name]["access_token"],
        api_base_url=my_config[group_name]["mastodon_instance"],
    )

    try:
        # Get the group account information.
        # This connects to the Mastodon or Pleroma server for the first time at
        # every tootgroup.py's run.
        my_account = {
            "username": masto.account_verify_credentials().username,
            "id": masto.account_verify_credentials().id,
            "group_member_ids": [],
        }
    except Exception as ex:
        print("")
        print("\n########################################################")
        print("tootgroup.py could not connect to the Mastodon or Pleroma")
        print("instance. If you know that it is running, there might be a")
        print("problem with your local configuration. Check the error")
        print("message for more details:")
        print(ex)
        print("\nYou can always try to delete tootgroup.py's config file")
        print("and re-run the script for a new setup.")
        print("########################################################\n")
        sys.exit(0)

    # Get group member IDs. They can not be fetched directly when
    # connecting to the Mastodon server.
    #
    # limit=sys.maxsize is set here because Pleroma only returns 20 Member if
    # the standard limit=None value is used!
    for member in masto.account_following(my_account["id"], limit=sys.maxsize):
        my_account["group_member_ids"].append(member.id)

    # Do we accept direct messages, public retoots, both or none? This
    # can be set in the configuration.
    accept_direct_messages = my_config[group_name].getboolean("accept_DMs")
    accept_retoots = my_config[group_name].getboolean("accept_retoots")
    dm_visibility = my_config[group_name]["dm_visibility"]

    # Get all new notifications up to a maximum number set here.
    # Defaults to 100 and has to be changed in the code if really desired.
    # Use the ID of the last seen notification for deciding what is new.
    notification_count = 0
    max_notifications = 100
    max_notification_id = None
    latest_notification_id = 0
    my_notifications = []
    get_more_notifications = True

    # Get notifications. Stop if either the last known notification ID or
    # the maximum number is reached. Chunk size is limited to 40 by default
    # in Mastodon but this can be configured at any specific server instance.
    while get_more_notifications:
        get_notifications = masto.notifications(max_id=max_notification_id)

        # Remember the ID of the latest notification on first iteration
        if notification_count == 0:
            if len(get_notifications) > 0:
                latest_notification_id = get_notifications[0].id
            else:  # If there have not been any notifications yet, set value to "0"
                latest_notification_id = 0
                print(
                    "Nothing to do yet! Start interacting with your group account first."
                )
                get_more_notifications = False
        else:  # leave the while loop if there are no more notifications fetched
            if len(get_notifications) == 0:
                get_more_notifications = False

        # Initialize "last_seen_id" on first run. Notifications are ignored up
        # to this point, but newer ones will be considered subsequently.
        if my_config[group_name]["last_seen_id"] == "catch-up":
            my_config[group_name]["last_seen_id"] = str(latest_notification_id)
            config_store["write_NEW"] = True
            print("Caught up to current timeline. Run again to start group-tooting.")

        for notification in get_notifications:
            max_notification_id = notification.id

            if (
                notification.id > int(my_config[group_name]["last_seen_id"])
                and notification_count < max_notifications
            ):
                my_notifications.append(notification)
            else:
                get_more_notifications = False
                break

            notification_count += 1

    # If there have been new notifications since the last run, update "last_seen_id"
    # and write config file to persist the new value.
    if latest_notification_id > int(my_config[group_name]["last_seen_id"]):
        my_config[group_name]["last_seen_id"] = str(latest_notification_id)
        config_store["write_NEW"] = True

    # Reverse list of notifications so that the oldest are processed first
    my_notifications = my_notifications[::-1]

    # run through the notifications and look for retoot candidates
    for notification in my_notifications:

        # Only from group members
        if notification.account.id in my_account["group_member_ids"]:

            # Is retooting of public mentions configured?
            if accept_retoots:
                if (
                    notification.type == "mention"
                    and notification.status.visibility == "public"
                ):
                    # Only if the mention was preceeded by either a "!" or a "*".
                    # To check for this, html tags have to be removed first.
                    repost_triggers = []
                    repost_triggers.append("!@" + my_account["username"])
                    repost_triggers.append("*@" + my_account["username"])
                    status = re.sub("<.*?>", "", notification.status.content)
                    if any(trigger in status for trigger in repost_triggers):
                        if not commandline_arguments["dry_run"]:
                            masto.status_reblog(notification.status.id)
                            print(
                                "Retooted from notification ID: " + str(notification.id)
                            )
                        else:
                            print(
                                "DRY RUN - would have retooted from notification ID: "
                                + str(notification.id)
                            )

            # Is reposting of direct messages configured? - if yes then:
            # Look for direct messages
            if accept_direct_messages:
                if (
                    notification.type == "mention"
                    and notification.status.visibility == "direct"
                ):
                    # Remove HTML tags from the status content but keep linebreaks
                    new_status = re.sub("<br />", "\n", notification.status.content)
                    new_status = re.sub("</p><p>", "\n\n", new_status)
                    new_status = re.sub("<.*?>", "", new_status)
                    # Get the group account's @username
                    account_username = "@" + my_account["username"]

                    # Only continue if the DM starts with the group account's @username
                    if new_status.startswith(account_username):
                        # Remove the group account's @username from the text
                        new_status = re.sub(account_username, "", new_status)
                        # "un-escape" HTML special characters
                        new_status = html.unescape(new_status)
                        if not commandline_arguments["dry_run"]:
                            # Repost as a new status
                            spoiler_text = notification.status.spoiler_text
                            # DMs from friendica always have a title that is used as spoiler_text.
                            # Remove it if it does not make sense.
                            if spoiler_text == "[no subject]":
                                spoiler_text = ""
                            masto.status_post(
                                new_status,
                                media_ids=media_toot_again(
                                    notification.status.media_attachments, masto
                                ),
                                sensitive=notification.status.sensitive,
                                visibility=dm_visibility,
                                spoiler_text=spoiler_text,
                            )
                            print(
                                "Newly posted from DM with notification ID: "
                                + str(notification.id)
                            )
                        else:
                            print(
                                "DRY RUN - would have newly posted from DM with notification ID: "
                                + str(notification.id)
                            )

                    # If the DM does not start with the group account's @username it will not
                    # trigger a new toot. Notify the originating user why this happened instead!
                    else:
                        if not commandline_arguments["dry_run"]:
                            new_status = (
                                "@" + notification.account.acct + "\n"
                                "Ohai! - This is a notification from your friendly tootgroup.py bot.\n\n"
                                "Your message has not been converted to a new group toot because it did "
                                "not start with @" + my_account["username"] + "\n\n"
                                "Remember to put @"
                                + my_account["username"]
                                + " at the very beginning if "
                                "you want to create a new group toot."
                            )
                            masto.status_post(
                                new_status,
                                in_reply_to_id=notification.status.id,
                                visibility="direct",
                            )
                            print(
                                "Not posted from DM with notification ID: "
                                + str(notification.id),
                                end="",
                            )
                            print(
                                ". It did not start with @"
                                + my_account["username"]
                                + "!"
                            )
                        else:
                            print(
                                "DRY RUN - received DM with notification ID: "
                                + str(notification.id)
                                + ", but it did not begin with @"
                                + my_account["username"]
                                + "!"
                            )

    # There have been changes requiring to persist the new configuration
    # but not in a dry-run condition
    if config_store["write_NEW"] and not commandline_arguments["dry_run"]:
        tootgroup_tools.configuration_management.write_configuration(
            config_store, my_config
        )

    print(
        "Successful tootgroup.py run for "
        + "@"
        + my_account["username"]
        + " at "
        + my_config[group_name]["mastodon_instance"]
    )


def media_toot_again(orig_media_dict, mastodon_instance):
    """Re-upload media files to the server for use in another toot.

    "orig_media_dict" - extracted media files from the original toot

    "mastodon_instance" - needed to re-upload the media files and create
    a new media_dict.

    Mastodon does not allow the re-use of already uploaded media files (images,
    videos) in a new toot. This function downloads all media files from a toot
    and re-uploads them. It then returns a dict formatted in a proper way to
    be used by the Mastodon.status_post() function.
    It also works with Pleroma this way altough it has not been tested if there
    would be another, more "direct" solution with that software."""
    new_media_dict = []
    for media in orig_media_dict:
        media_data = requests.get(media.url).content
        filename = os.path.basename(media.url)
        # basename still includes a "?" followed by a number after the file's name.
        # Remove them both.
        filename = filename.split("?")[0]
        # separate filename and extension
        filename, file_ext = os.path.splitext(filename)
        f_temp = tempfile.NamedTemporaryFile(suffix=file_ext, delete=False, mode="w+b")
        try:
            f_temp.write(media_data)
            f_temp.close()  # close to flush data
            # This re-uploads the current media file to the server!
            new_media_dict.append(
                mastodon_instance.media_post(f_temp.name, description=media.description)
            )
        except Exception as ex:
            print("")
            print(
                "\n##################################################################"
            )
            print("Cannot write media to temporary file:")
            print(ex)
            print("")
            print("tootgroup.py will continue but media files might not get reposted!")
            print(
                "##################################################################\n"
            )
        finally:
            f_temp.close()  # close again for good measure
            try:
                os.unlink(f_temp.name)  # delete temporary file
            except Exception:
                pass  # cannot delete, probably non-existant anyway!
    return new_media_dict


# Start executing main() function if the script is called from a command line
if __name__ == "__main__":
    main()
